from __future__ import annotations
import typing as t

import pydantic
import pydantic_core

__all__ = (
    "SpecialStrBase",
)


# coercible type annotation specifying you could also initialize this class with.
_CoercibleTV = t.TypeVar("_CoercibleTV", )
SelfTV = t.TypeVar("SelfTV", bound="SpecialStrBase")


class SpecialStrBase(str):
    """container for a special string to avoid having """

    @classmethod
    def _parse_non_str(cls, root_data: t.Any) -> str:
        """overwrite this to allow the special str class to parse non-string input"""
        raise TypeError(f"'{cls.__name__}' cannot parse data of type {type(root_data)}")

    @classmethod
    def _format_str_val(cls, root_data: str) -> str:
        """overwrite this to specify the custom str formatting for this class"""
        return root_data

    @classmethod
    def _validate_str_val(cls, root_data: str) -> str:
        """validate that the formatted string data is valid"""
        return root_data

    @classmethod
    def _validate(cls, val: t.Any) -> str:
        str_val: str
        if not isinstance(val, str):
            str_val = cls._parse_non_str(val)
        else:
            str_val = val
        str_val = cls._format_str_val(str_val)
        return cls._validate_str_val(str_val)

    def __new__(cls, v: t.Union[str, _CoercibleTV]):
        """create a new instance of this special type class, validating the data first"""
        v = cls._validate(v)
        return super().__new__(cls, v)

    @classmethod
    def __get_pydantic_core_schema__(
            cls, _: t.Any, handler: pydantic.GetCoreSchemaHandler
    ) -> pydantic_core.CoreSchema:
        final_schema = pydantic_core.core_schema.no_info_after_validator_function(
            function=cls, schema=handler(str)
        )
        return final_schema

    def special_repr(self) -> str:
        """
        Gives us a special representation indicating the instance is a special string, not just
        a normal one.

        This keeps the regular '__repr__' method is the same as a normal str instance
        """
        return f"{self.__class__.__name__}['{str(self)}']"

    def __eq__(self: SelfTV, __value: object) -> bool:
        try:
            other_formatted = self._validate(__value)
            return str(self) == str(other_formatted)

        except Exception:
            return False



from __future__ import annotations
import typing as t
import json
import abc

import pydantic


__all__ = (
    "BASE_MODEL_CONFIG",
    "CommonModelMethods"
)

BASE_MODEL_CONFIG = pydantic.ConfigDict(
    extra="forbid",
)


SelfTV = t.TypeVar("SelfTV", bound=pydantic.BaseModel)


class CommonModelMethods(abc.ABC):
    """a base model we use for all pydantic models, even the other bases"""
    def __iter__(self) -> t.Any:
        raise NotImplementedError(
            f"'{self.__class__.__name__}' doesn't have iteration implemented"
        )

    def json_dict(
            self,
            use_str_fallback: bool = False,
            **pydantic_kwargs
    ) -> dict:
        """
        Converts a pydantic instance to a dict, going through json first.
        This is a convenience function for when you want a dict, but want to use the json
        encoders connected to the pydantic object.

        Args:
            use_str_fallback: if true, we cast anything we cannot encode to a string
            **pydantic_kwargs: any kwargs available for pydantic's json method. see their docs

        Returns:
            dict: a natively json-encodable dict

        Notes:
            If the obj is a RootModel, calling `obj.json` would not necessarily return a dict,
            so we convert it to the form, `{"__root__": data}` to align with how pydantic
            would return `obj.dict()` in this scenario.

        Examples:
            simple class with custom date encoder::

                class Data(PydanticBase):
                    class Config:
                        json_encoders = {
                            dt.date: lambda date_obj: date_obj.strftime("%d/%m/%Y")
                        }
                    x: int
                    date: dt.date

                inst = Data(x=1, date=dt.date(year=2023, month=1, day=1))

                # regular dict
                inst.dict() == {"x": 1, "date": dt.date(year=2023, month=1, day=1)}

                # this method
                inst.json_dict() == {"x": 1, "date": "01/01/2023"}
                # note that the date is the json encoding tied to the encoder for this pydantic
                # class. I just used a familiar string format, it could be whatever is specified
        """
        pydantic_kwargs = pydantic_kwargs if pydantic_kwargs is not None else {}
        if use_str_fallback:
            fallback = str
        else:
            fallback = None

        pydantic_model_inst: pydantic.BaseModel = t.cast(
            pydantic.BaseModel, self
        )

        pydantic_model_inst.model_dump_json()
        dumped_json = pydantic_model_inst.__pydantic_serializer__.to_json(
            self,
            fallback=fallback,
            **pydantic_kwargs,
        )
        return json.loads(dumped_json)

    @classmethod
    def _import_forward_refs(cls) -> dict:
        """
        hook provided for convenience when you define a subclass with forward refs,
        to avoid circular import issue. Overwrite this method to do the imports and
        add the newly imported objects to a dict.

        Notes:
            - If you're not sure what to import and what keys to use, look at the stuff
              in the `if TYPE_CHECKING:` block, paste those exact imports in here and
              set them equal to their variable name within your current module's context.

        Returns:
            dict: a dict of references to update

        """
        return {}

    @classmethod
    def update_forward_refs(cls, **kwargs) -> None:
        """
        This extends pydantic's builtin hook for updating forward refs.
        We call our special '_import_forward_refs' hook to automatically bring in the things
        imported there, but also if you want to use it like pydantic, i.e. call it directly and
        provide your refs to update via kwargs, that is still supported.

        Returns:
            None: doesn't return anything, mutates the class calling this method

        Raises:
            NameError: pydantic raises a NameError if a forward ref is not accounted for
             in the kwargs
        """
        imported_refs = cls._import_forward_refs()
        kwargs.update(imported_refs)
        return super().update_forward_refs(**kwargs)  # type: ignore



from __future__ import annotations
import typing as t

from spice_rack._logging._logger import Logger


__all__ = (
    "get_logger",
)


def get_logger(
        specified_service_name: t.Optional[str] = None,
        persist_service_name: bool = False
) -> Logger:
    logger_inst = Logger(service_name=specified_service_name)
    if persist_service_name:
        logger_inst.cache_service_name()
    return logger_inst



from spice_rack._bases import (
    _base_base as base_base,
    _value_model as value_model,
    _dispatchable as dispatchable,
    _special_str as special_str,
    _settings as settings,
    _exception as exceptions
)



from __future__ import annotations
import typing as t

import typing_extensions as t_ext
import pydantic
import inflection

from spice_rack._bases import _special_str
from spice_rack._bases import _base_base


__all__ = (
    "DispatchedModelMixin",
    "ClassId",
    "DispatchedClassContainer"
)


class ClassId(_special_str.SpecialStrBase):
    """a class id for a member of the dispatched family"""

    @classmethod
    def _format_str_val(cls, root_data: str) -> str:
        return inflection.underscore(root_data)


TypeTV = t.TypeVar("TypeTV", bound=ClassId)
SelfTV = t.TypeVar("SelfTV", bound="DispatchedModelMixin")

_LiteralT: t_ext.TypeAlias = type(t.Literal["xxx"])  # type: ignore

ClassMetaTypeT = t.Literal["root", "base", "concrete"]


def _json_schema_extra(schema: dict) -> None:
    """remove default from class id field and add it to the required fields"""
    schema["properties"]["class_id"].pop("default", None)
    required_fields = schema.get("required", [])
    if "class_id" not in required_fields:
        required_fields = ["class_id", *required_fields]
    schema["required"] = required_fields


_model_config = _base_base.BASE_MODEL_CONFIG
_model_config["json_schema_extra"] = _json_schema_extra


class DispatchedModelMixin(pydantic.BaseModel, _base_base.CommonModelMethods, t.Generic[TypeTV]):
    """this creates a root of this dispatched class."""
    model_config = _model_config

    _cls_id: t.ClassVar[ClassId]
    _cls_meta_type: t.ClassVar[ClassMetaTypeT]
    _parent_cls_id_path: t.ClassVar[tuple[ClassId, ...]] = ()

    class_id: TypeTV = pydantic.Field(
        description="this will be overwritten in concrete classes to t.Literal['<cls_id_str>]', "
                    "and be str in base classes",
        default=None,
    )

    @classmethod
    def model_parametrized_name(cls, params: tuple[type[t.Any], ...]) -> str:
        res = super().model_parametrized_name(params)
        return f"_ParameterizedPassthrough{res}"

    def __class_getitem__(
            cls: t.Type[SelfTV],
            params: t.Union[
                t.TypeVar,
                _LiteralT,
                t.Tuple[
                    str,
                    ClassMetaTypeT,
                ]
            ]
    ) -> t.Type[SelfTV]:
        """
        the bracket style subclassing

        Args:
            params: todo: doc each type of arg

        Returns:
            a newly built class
        """
        res: t.Type[SelfTV]

        # this is the base situation so keep passing back
        if t.get_origin(params) == t.Literal:
            res = super().__class_getitem__(params)  # type: ignore

        # this is a root class, i.e. it a new parent
        elif isinstance(params, t.TypeVar):
            res = super().__class_getitem__(params)  # type: ignore
            res._cls_id = ClassId(res.__name__)

        else:
            cls_id = ClassId(params[0])
            cls_meta_type = params[1]
            if cls_meta_type == "concrete":
                literal_t = t.Literal[(str(cls_id),)]  # type: ignore
                res = super().__class_getitem__(literal_t)  # type: ignore
                model_field = res.model_fields["class_id"]
                model_field.default = str(cls_id)

            else:
                res = super().__class_getitem__(TypeTV)  # type: ignore

            res._cls_id = cls_id
            res._cls_meta_type = cls_meta_type
            res._parent_cls_id_path += (ClassId(cls.get_class_type()), )
        return res

    @classmethod
    def get_class_meta_type(cls) -> ClassMetaTypeT:
        return cls._cls_meta_type
    
    @classmethod
    def get_class_type(cls) -> str:
        if not hasattr(cls, "_cls_id"):
            raise ValueError(
                f"'{cls.__name__}' didn't set their '_cls_id' class attribute"
            )
        else:
            return str(cls._cls_id)

    @classmethod
    def _is_parameterized_passthrough(cls) -> bool:
        return cls.__name__.startswith("_ParameterizedPassthrough")

    @classmethod
    def is_concrete(cls) -> bool:
        return cls._cls_meta_type == "concrete" and not cls._is_parameterized_passthrough()

    @classmethod
    def iter_concrete_subclasses(cls: t.Type[SelfTV]) -> t.Iterator[t.Type[SelfTV]]:
        if cls.__name__ == DispatchedModelMixin.__name__:
            raise ValueError(
                "don't call iter_concrete_subclasses on the DispatchedModelMixin class "
                "itself, only on root classes built from it."
            )
        if cls.is_concrete():
            yield cls
        else:
            for sub_cls in cls.__subclasses__():
                for sub_sub_cls in sub_cls.iter_concrete_subclasses():
                    yield sub_sub_cls

    @classmethod
    def model_validate(
        cls: t.Type[SelfTV],
        obj: t.Any,
        *args,
        **kwargs,
    ) -> SelfTV:
        if cls._cls_meta_type != "concrete":
            raise ValueError(
                f"'{cls.get_class_type()}' is not concrete, so we do not want"
                f" to initialize instances of this class"
            )
        else:
            return t.cast(SelfTV, super().model_validate(obj, *args, **kwargs))

    @classmethod
    def build_dispatched_ann(cls: t.Type[SelfTV]) -> t.Type[SelfTV]:
        options = list(cls.iter_concrete_subclasses())

        if len(options) == 0:
            raise ValueError(
                f"'{cls.__name__}' has no concrete options"
            )
        if len(options) > 1:
            union_t = t.Union[tuple(options)]  # type: ignore
            annotated_t = t.Annotated[
                union_t,  # type: ignore
                pydantic.Discriminator("class_id")
            ]
            return annotated_t  # type: ignore
        else:
            return options[0]

    @classmethod
    def build_dispatcher_type_adapter(cls: t.Type[SelfTV]) -> pydantic.TypeAdapter[SelfTV]:
        return pydantic.TypeAdapter(cls.build_dispatched_ann())


DispatchedClsTV = t.TypeVar("DispatchedClsTV", bound=DispatchedModelMixin)


class DispatchedClassContainer(pydantic.RootModel[DispatchedClsTV], t.Generic[DispatchedClsTV]):
    root: DispatchedClsTV = pydantic.Field(
        description="the root class ",
    )

    def __class_getitem__(
            cls,
            item: t.Type[DispatchedModelMixin]  # type: ignore
    ) -> t.Type[DispatchedClsTV]:
        if not issubclass(item, DispatchedModelMixin):
            raise ValueError(f"item must be subclass of DispatchedModelMixin, but {item} is not")
        return super().__class_getitem__(
            item.build_dispatched_ann()
        )  # type: ignore



from __future__ import annotations
import typing as t
import datetime as dt
import dateparser
import pydantic

from spice_rack._timestamp._tz_key import TimeZoneKey
from spice_rack import _logging


__all__ = (
    "Timestamp",
)


_TzKeyT = t.Union[str, TimeZoneKey, t.Literal["local"]]
_TimestampInitValueT = t.Union[float, str, dt.datetime, dt.date]
_PythonTimestampT = float  # seconds from epoch with decimals as sub-second measurements


class Timestamp(pydantic.RootModel[int], _logging.log_extra.LoggableObjMixin):
    """
    special subclass of 'int' that contains the utc millisecond from epoch.
    """
    _default_assumed_tz: t.ClassVar[TimeZoneKey] = TimeZoneKey("UTC")
    """timezone we assume when parsing raw data into this object"""
    # _default_tz: ClassVar[TimeZoneKey] = TimeZoneKey.local()

    def to_python_timestamp(self) -> _PythonTimestampT:
        # convert to seconds from epoch, posix timestamp
        return float(self.root) / 1000

    def to_dt_obj(
            self,
            with_tz: t.Optional[_TzKeyT] = None
    ) -> dt.datetime:
        res: dt.datetime

        utc_dt_obj = dt.datetime.fromtimestamp(
            self.to_python_timestamp(), tz=TimeZoneKey("UTC").as_zone_info()
        )
        if with_tz:
            tz_key: TimeZoneKey
            if with_tz == "local":
                tz_key = TimeZoneKey.local()
            else:
                tz_key = TimeZoneKey(str(with_tz))

            new_tz_info = tz_key.as_zone_info()
            obj_new_tz = utc_dt_obj.astimezone(tz=new_tz_info)
            res = obj_new_tz
        else:
            res = utc_dt_obj
        return res

    def to_iso_str(
            self,
            with_tz: t.Optional[_TzKeyT] = None
    ) -> str:
        """
        returns the datetime object as a formatted str

        Args:
            with_tz: a timezone key we can specify. use 'local' to use whatever timezone is local
                to the process calling this function
        Returns:
            str: formatted string
        """
        dt_obj = self.to_dt_obj(with_tz=with_tz)
        return dt_obj.isoformat()

    def to_pretty_str(
            self,
            fmat: str = "%x %X %Z",
            with_tz: t.Optional[_TzKeyT] = "local"
    ) -> str:
        """
        returns the datetime object as a formatted str, in a pretty, super readable format.
        The default is "%c %Z"

        Args:
            fmat: Optional custom formatting. See python datetime docs for formatting help.
                If not specified, we use 'auto' selected appropriate format.

            with_tz: a timezone key we can specify. use 'local' to use whatever timezone is local
                to the process calling this function
        Returns:
            str: formatted string

        Notes:
            - see the docs on format codes:
              https://docs.python.org/3/library/datetime.html#strftime-and-strptime-format-codes
        """
        dt_obj = self.to_dt_obj(with_tz=with_tz)
        return dt_obj.strftime(fmat)

    def special_repr(self, with_tz: t.Optional[_TzKeyT] = None) -> str:
        return f"{self.__class__.__name__}['{self.to_iso_str(with_tz=with_tz)}']"

    @classmethod
    def _from_datetime_obj(
            cls,
            value: dt.datetime,
            assumed_tz: TimeZoneKey
    ) -> int:
        tz_info: TimeZoneKey
        if value.tzinfo:
            tz_info = TimeZoneKey(value.tzname())
        else:
            tz_info = assumed_tz

        dt_obj_utc = value.astimezone(tz_info.as_zone_info())
        return int(dt_obj_utc.timestamp() * 1000)

    @classmethod
    def _from_date_obj(
            cls,
            value: dt.date,
            assumed_tz: TimeZoneKey
    ) -> int:
        datetime_obj = dt.datetime.fromisoformat(value.isoformat())
        return cls._from_datetime_obj(datetime_obj, assumed_tz=assumed_tz)

    @classmethod
    def _from_str(
            cls,
            value: str,
            assumed_tz: TimeZoneKey
    ) -> int:
        """parse a str to datetime object then to timestamp"""
        dt_obj_maybe = dateparser.parse(value)
        if dt_obj_maybe is None:
            raise ValueError(
                f"cannot parse the string, '{value}', to a datetime object "
            )

        else:
            dt_obj: dt.datetime = dt_obj_maybe
        return cls._from_datetime_obj(value=dt_obj, assumed_tz=assumed_tz)

    @classmethod
    def _from_float(
            cls,
            value: float,
            assumed_tz: TimeZoneKey
    ) -> int:
        raise ValueError("initializing from a float is not currently supported")

    @pydantic.model_validator(mode="before")
    def _validate(cls, value: _TimestampInitValueT) -> int:
        int_data: int

        if isinstance(value, str):
            int_data = cls._from_str(value, assumed_tz=cls._default_assumed_tz)

        elif isinstance(value, dt.datetime):
            int_data = cls._from_datetime_obj(value, assumed_tz=cls._default_assumed_tz)

        elif isinstance(value, dt.date):
            int_data = cls._from_date_obj(value, assumed_tz=cls._default_assumed_tz)

        # elif isinstance(value, float):
        #     float_data = cls._from_float(value, assumed_tz=cls._default_assumed_tz)

        elif isinstance(value, (int, cls)):
            int_data = value

        else:
            raise ValueError(
                f"'{cls.__name__}' doesn't support data of type {type(value)}"
            )

        return int_data

    @classmethod
    def now(cls) -> Timestamp:
        """get now down to the millisecond, not microsecond"""
        return Timestamp(dt.datetime.utcnow())

    @classmethod
    def today(cls) -> Timestamp:
        """get the current date, as a Timestamp object."""
        dt_obj = dt.datetime.today()
        return Timestamp(dt_obj)

    def format_for_logger(self) -> _logging.log_extra.ExtraLogData:
        return _logging.log_extra.ExtraLogData(
            key="timestamp",
            desc="a timestamp object",
            data=self.to_iso_str()
        )
